if not (rawget(_G, "customhud"))
	return;
end

local modname = "sa1hud";

customhud.SetupNumberFont("SANUM", -1, 4);
customhud.SetupNumberFont("SARNUM", -1, 4);

customhud.SetupNumberFont("SALNUM", -1, 4);
customhud.SetupFont("SAFNT", -1, 4);
customhud.SetupFont("SASFT", 0, 4);

customhud.SetupFont("SABFT", -1, 4);

customhud.SetupNumberFont("GMNUM", -1, 4);
customhud.SetupNumberFont("GMSNUM", -1, 4);
customhud.SetupNumberFont("GMWNUM", -1, 4);
customhud.SetupNumberFont("GMRNUM", -1, 4);
customhud.SetupFont("GMSFT", -1, 4);

local dontshow = "_DONTSHOWTHISBOSSMETER"; -- hopefully no one makes a boss named this? :p
local cardlength = (2*TICRATE) + (NUMTRANSMAPS*2);

local bossnames = {
	-- Vanilla SRB2
	[MT_EGGMOBILE] = "Egg Mobile",
	[MT_EGGMOBILE2] = "Egg Slimer",
	[MT_EGGMOBILE3] = "Sea Egg",
	[MT_EGGMOBILE4] = "Egg Coliseum",
	[MT_FANG] = "Fang",
	[MT_METALSONIC_BATTLE] = "Metal Sonic",
	[MT_CYBRAKDEMON] = "Black Eggman",
	[MT_BLACKEGGMAN] = "Black Eggman",
};

local addonbosses = {
	-- SUGOI
	MT_SILVER_318 = "Silver",
	MT_TURRETMOBILE = "Turret Mobile",
	MT_PURPLEHEART = dontshow,
	-- SUBARASHII
	MT_EGGDOMINATOR = "Egg Dominator",
	MT_RRZBOSS = dontshow,
	MT_EGGWORLD = dontshow,
	MT_EXPSTAR = dontshow,
	-- KIMOKAWAIII
	MT_EGGGUNNER = "Egg Gunner",
	MT_BOUNCYBOI = "BoBo the Bouncy Boi",
	MT_CYBERDETON = "Cyber Deton",
	MT_POINTYEGG = "Pointy Egg",
	MT_LJSONIC = "LJ Sonic",
	MT_CONFLEGGRATOR = "Confleggrator",
	MT_AALPBOSS = "Egg Robo",
	MT_DARKSTABOT = "Dark Stabot",
	MT_EGGLAS = "Egg Lasershow",
	MT_PROTAGONIST = "True Protagonist",
	MT_GUNDAM = dontshow,
	MT_SOAES = dontshow,
	-- (potential) character bosses
	MT_SONIC = "Sonic",
	MT_TAILS = "Tails",
	MT_KNUCKLES = "Knuckles",
	MT_AMY = "Amy Rose",
	MT_SHADOW = "Shadow",
	MT_SILVER = "Silver",
	-- Misc levels
	MT_EGGMOBILE7 = "Egg Boiler",
	MT_STRAYBOLTS_BOSS = "Stray-Bolts",
	MT_OLDK = "K.T.E",
	MT_ANASTASIA = dontshow,
	MT_BOSSRIDE = "Player", -- this one never gets used, included anyway to tell the game you have it loaded
	-- True Arena's ports
	MT_SANDSUB_326 = "Egg Sand Sub",
	MT_GREENHILLBOSS = "Ball & Chain",
	MT_SUPERHOOD = "Super Robo-Hood",
	MT_INFINITE_318 = "Infinite",
};

-- get around stupid "can only be used in a level" warnings
local function isFuckingSpecialStage(mapnum)
	if (mapnum == nil)
		mapnum = gamemap;
	end

	if (modeattacking == ATTACKING_RECORD)
		return false;
	end

	if (mapnum >= sstage_start and mapnum <= sstage_end)
		return true;
	end

	if (mapnum >= smpstage_start and mapnum <= smpstage_end)
		return true;
	end

	return false;
end

local function gametypeFuckingUsesLives()
	if ((gametyperules & GTR_LIVES)
	and not (modeattacking or metalrecording)
	and not isFuckingSpecialStage()
	and not (maptol & TOL_NIGHTS))
		return true;
	end

	return false;
end

local function isMobjTypeValid(mt)
	if (pcall(do return _G[mt] end))
		return _G[mt];
	else
		return nil;
	end
end

local function retrieveBossName(boss)
	local name = "Target";

	if (bossnames[boss.type])
		name = bossnames[boss.type];

		if (isMobjTypeValid("MT_BOSSRIDE"))
			if (_G["MT_BOSSRIDE"] == boss.type)
			and ((boss.rider and boss.rider.valid)
			and (boss.rider.player and boss.rider.player.valid))
				name = boss.rider.player.name;
			end
		end
	end

	return name;
end

local function setBossMeter(player, boss)
	if (player.sabossmeter != nil)
	and (player.sabossmeter.mo == boss)
		-- It's the same guy!
		return;
	end

	local name = retrieveBossName(boss);
	if (name == dontshow)
		-- We don't want a meter for this boss, keep the old one
		return;
	end

	if (player.sabossmeter == nil)
		player.sabossmeter = {};
		player.sabossmeter.barlength = 0;
		player.sabossmeter.disphp = 0;
	end

	player.sabossmeter.mo = boss;
	player.sabossmeter.name = name;
end

local function bossThink(mo)
	if (mo.health <= 0)
		-- Don't add a boss that's dead!
		return;
	end

	-- I suppose we can iterate everyone...
	for player in players.iterate do
		if not (player.mo and player.mo.valid)
			-- It's already us! We can move on...
			continue;
		end

		if (player.sabossmeter != nil
		and ((player.sabossmeter.mo and player.sabossmeter.mo.valid)
		and player.sabossmeter.mo == mo))
			-- It's already us! We can move on...
			continue;
		end

		if (P_CheckSight(mo, player.mo))
			local updateboss = false;

			if ((player.sabossmeter == nil)
			or not (player.sabossmeter.mo and player.sabossmeter.mo.valid and player.sabossmeter.mo.health > 0))
				-- Another boss doesn't exist, so we can add it!
				updateboss = true;
			else
				-- Another boss exists already, so only assume that they're not fighting the boss when
				updateboss = (not P_CheckSight(player.sabossmeter.mo, player.mo));
			end

			if (updateboss == true)
				setBossMeter(player, mo);
			end
		end
	end
end

local function bossHurt(mo, inf, src)
	if not (mo.flags & MF_BOSS)
		-- Only bosses!
		return;
	end

	if (src and src.valid) and (src.player and src.player.valid)
		setBossMeter(src.player, mo);
	end
end

local function bossMeterThink(player)
	if (player.sabossmeter == nil)
		return;
	end

	local wantedhp = 0;

	if (player.sabossmeter.barlength > 0)
	and not (player.sabossmeter.mo and player.sabossmeter.mo.valid and player.sabossmeter.mo.health > 0)
		player.sabossmeter.barlength = $1-2;
	else
		local maxhealth = player.sabossmeter.mo.info.spawnhealth;
		local seglength = 12;

		if (isMobjTypeValid("MT_OLDK"))
			if (_G["MT_OLDK"] == boss.type)
				seglength = $1 * 8; -- Give KTE a full-sized health bar!
				maxhealth = 1; -- (why does he have 2 spawn health despite only being 1 hit though)
			end
		end

		player.sabossmeter.barlength = maxhealth*seglength;

		while ((player.sabossmeter.barlength > 168) and (seglength > 1))
			seglength = $1-1;
			player.sabossmeter.barlength = maxhealth*seglength;
		end

		local bosshp = player.sabossmeter.mo.health * seglength;
		if (player.sabossmeter.mo.raidBoss != nil)
			local extraDmg = FRACUNIT - player.sabossmeter.mo.raidBoss.poise;
			bosshp = $1 - FixedMul(seglength, extraDmg);
			if (player.sabossmeter.mo.health > 0)
				bosshp = max(bosshp, 1);
			end
		end
		wantedhp = min(player.sabossmeter.barlength, bosshp);
	end

	if (player.sabossmeter.barlength <= 0)
		player.sabossmeter = nil;
		return;
	end

	if (player.sabossmeter.disphp != wantedhp)
		local dir = -1;
		if (player.sabossmeter.disphp < wantedhp)
			dir = 1;
		end

		player.sabossmeter.disphp = $1 + dir;
	end

	if (player.sabossmeter.disphp > player.sabossmeter.barlength)
		player.sabossmeter.disphp = player.sabossmeter.barlength;
	end
end

addHook("BossThinker", bossThink);
addHook("MobjDamage", bossHurt);

addHook("ThinkFrame", function()
	for player in players.iterate do
		bossMeterThink(player);
	end
end);

local intermissionData = nil;
local prevScore = 0;

local function timeBonus(player)
	local bonus = nil;
	local secs = player.realtime / TICRATE;

	if     (secs <  30) bonus = 50000;
	elseif (secs <  60) bonus = 10000;
	elseif (secs <  90) bonus = 5000;
	elseif (secs < 120) bonus = 4000;
	elseif (secs < 180) bonus = 3000;
	elseif (secs < 240) bonus = 2000;
	elseif (secs < 300) bonus = 1000;
	elseif (secs < 360) bonus = 500;
	elseif (secs < 420) bonus = 400;
	elseif (secs < 480) bonus = 300;
	elseif (secs < 540) bonus = 200;
	elseif (secs < 600) bonus = 100;
	else bonus = 0; end

	return {
		title = "Time score",
		value = bonus,
	};
end

local function ringBonus(player)
	return {
		title = "Ring score",
		value = max(0, player.rings * 100),
	};
end

local function nightsBonus(player)
	return {
		title = "NiGHTS score",
		value = max(0, player.totalmarescore),
	};
end

local function lapBonus(player)
	return {
		title = "Lap score",
		value = max(0, player.totalmarebonuslap * 1000),
	};
end

local function linkBonus(player)
	return {
		title = "Link score",
		value = max(0, (player.maxlink - 1) * 100),
	};
end

local function guardBonus(player)
	local bonus = nil;

	if     (player.timeshit == 0) bonus = 10000;
	elseif (player.timeshit == 1) bonus = 5000;
	elseif (player.timeshit == 2) bonus = 1000;
	elseif (player.timeshit == 3) bonus = 500;
	elseif (player.timeshit == 4) bonus = 100;
	else bonus = 0; end

	return {
		title = "Guard score",
		value = bonus,
	};
end

--[[
local function perfectBonus(player)
	-- annoying to do without nummaprings exposed
end
--]]

local function specialRingBonus(player)
	local bonus = nil;
	local sharedRings = 0;

	for p in players.iterate do
		sharedRings = $1 + p.rings;
	end

	return {
		title = "Ring score",
		value = max(0, (sharedRings) * 100),
	};
end

local function createBonusList(player)
	local bonusType = 0;

	if (player.lives < 1)
		bonusType = 0;
	elseif (isFuckingSpecialStage())
		if (maptol & TOL_NIGHTS)
			bonusType = -1;
		else
			bonusType = -2;
		end
	elseif (mapheaderinfo[gamemap] != nil)
		bonusType = mapheaderinfo[gamemap].bonustype + 1;
	end

	local numBonuses = 1;
	local bonuses = {};

	bonuses[numBonuses] = {
		title = "Stage score",
		value = 0, -- Updated later
	};

	if (bonusType == 1)
		numBonuses = $1 + 1;
		bonuses[numBonuses] = timeBonus(player);

		numBonuses = $1 + 1;
		bonuses[numBonuses] = ringBonus(player);

		--numBonuses = $1 + 1;
		--bonuses[numBonuses] = perfectBonus(player);
	elseif (bonusType == 2)
		numBonuses = $1 + 1;
		bonuses[numBonuses] = guardBonus(player);

		numBonuses = $1 + 1;
		bonuses[numBonuses] = ringBonus(player);
	elseif (bonusType == 3)
		numBonuses = $1 + 1;
		bonuses[numBonuses] = guardBonus(player);

		numBonuses = $1 + 1;
		bonuses[numBonuses] = ringBonus(player);

		--numBonuses = $1 + 1;
		--bonuses[numBonuses] = perfectBonus(player);
	elseif (bonusType == 4)
		numBonuses = $1 + 1;
		bonuses[numBonuses] = nightsBonus(player);

		numBonuses = $1 + 1;
		bonuses[numBonuses] = lapBonus(player);
	elseif (bonusType == 5)
		numBonuses = $1 + 1;
		bonuses[numBonuses] = linkBonus(player);

		numBonuses = $1 + 1;
		bonuses[numBonuses] = lapBonus(player);
	elseif (bonusType == -1) -- Special stage bonuses
		numBonuses = $1 + 1;
		bonuses[numBonuses] = nightsBonus(player);
	elseif (bonusType == -2)
		numBonuses = $1 + 1;
		bonuses[numBonuses] = specialRingBonus(player);

		--numBonuses = $1 + 1;
		--bonuses[numBonuses] = perfectBonus(player);
	end

	return bonuses;
end

local function intThink(player)
	if (intermissionData == nil)
		-- Initalize
		intermissionData = {};

		intermissionData.pause = TICRATE;
		intermissionData.displayValues = 0;
		intermissionData.buttonHeld = true;

		intermissionData.time = player.realtime;
		intermissionData.rings = player.rings;
		intermissionData.stageScore = player.score - prevScore;

		intermissionData.bonusList = createBonusList(player);
		intermissionData.total = 0;

		intermissionData.finished = false;

		local nb = #(intermissionData.bonusList);
		if (nb > 1)
			for i = 2,nb do
				local bonusData = intermissionData.bonusList[i];
				intermissionData.stageScore = $1 - bonusData.value;
			end
		end

		if (intermissionData.stageScore < 0)
			intermissionData.stageScore = 0;
		end

		intermissionData.bonusList[1].value = intermissionData.stageScore;

		prevScore = player.score;
	end

	if (intermissionData.finished == true)
		return;
	end

	local numBonuses = #(intermissionData.bonusList);

	// (time + rings + total) + bonuses
	local maxValues = 3 + numBonuses;

	if (player.cmd.buttons & BT_SPIN)
		-- Instantly finish
		intermissionData.pause = 0;
		intermissionData.displayValues = maxValues;

		for i = 1,numBonuses do
			local bonusData = intermissionData.bonusList[i];

			if (bonusData.value <= 0)
				continue;
			end

			intermissionData.total = $1 + bonusData.value;
			bonusData.value = 0;
		end

		S_StartSound(nil, sfx_s3kb0, player);
		intermissionData.finished = true;
		return;
	end

	if (intermissionData.pause > 0)
		intermissionData.pause = $1 - 1;
		return;
	end

	if (intermissionData.displayValues < maxValues)
		intermissionData.displayValues = $1 + 1;
		intermissionData.pause = TICRATE/5;
		S_StartSound(nil, sfx_s3kb7, player);
		return;
	end

	local allDone = true;
	for i = 1,numBonuses do
		local bonusData = intermissionData.bonusList[i];

		if (bonusData.value <= 0)
			continue;
		end

		local reduceVal = 444; --222

		if (bonusData.value > reduceVal)
			intermissionData.total = $1 + reduceVal;
			bonusData.value = $1 - reduceVal;
			allDone = false;
		else
			intermissionData.total = $1 + bonusData.value;
			bonusData.value = 0;
		end
	end

	if (allDone == true)
		intermissionData.finished = true;
		S_StartSound(nil, sfx_s3kb0, player);
	else
		S_StartSound(nil, sfx_s3k5b, player);
	end
end

addHook("IntermissionThinker", function()
	if (gametype != GT_COOP)
		return;
	end

	if not (consoleplayer and consoleplayer.valid)
		return;
	end

	intThink(consoleplayer);
end);

local function mapSet(mo)
	-- Check for new addon bosses
	for k,v in pairs(addonbosses) do
		local mt = isMobjTypeValid(k);

		if not (mt)
			continue;
		end

		bossnames[mt] = v;
	end

	-- Uninit boss meters
	for player in players.iterate do
		player.sabossmeter = nil;
	end

	-- Reset intermission data
	intermissionData = nil;
end

addHook("MapChange", mapSet);

local screenwidth = 320;
local screenheight = 200;

local function sadxTime(v, player)
	if (player.bot)
		return;
	end

	if (maptol & TOL_NIGHTS) or (isFuckingSpecialStage())
		return;
	end

	if not (player and player.valid) return; end

	local startx = 8;
	local starty = 12;

	local label = v.cachePatch("SATIME");
	local colon = v.cachePatch("SACOLON");

	local displayTime = player.realtime;
	local pause = 0;
	local countdown = false;

	local levelTimeLeft = nil;

	if (mapheaderinfo[gamemap].countdown > 0)
		// do your best to barely support this instance
		levelTimeLeft = (mapheaderinfo[gamemap].countdown * TICRATE);
		displayTime = max(0, levelTimeLeft - leveltime);
		countdown = true;
	end

	if (player.gammaTimerRan != nil)
		if (player.gammaTime != nil)
			if (countdown == true)
				displayTime = min($1, player.gammaTime);
			else
				displayTime = player.gammaTime;
				pause = player.gammaTimePause;
				countdown = true;
			end
		elseif (countdown == false)
			displayTime = player.gammaTimeAttack;
			pause = player.gammaTimeAttackPause;
		end
	end

	local maxTime = (100 * 60 * TICRATE) - 1;
	displayTime = min($1, maxTime);

	local flags = V_SNAPTOLEFT|V_SNAPTOTOP|V_PERPLAYER;
	v.draw(startx, starty, label, flags|V_HUDTRANS);

	if (pause != 0)
	and (pause / 10 & 1)
		// Draw attention to the time limit by flashing it
		return;
	end

	if (displayTime < 30*TICRATE) and (countdown == true)
		local origvalue = v.localTransFlag() >> V_ALPHASHIFT;
		local length = NUMTRANSMAPS - origvalue;
		local pulse = abs(((leveltime/2) % (NUMTRANSMAPS*2)) - NUMTRANSMAPS);
		local trans = 0;

		if (origvalue >= NUMTRANSMAPS)
			return;
		end

		if (origvalue > 0)
			pulse = ($1 * length) / NUMTRANSMAPS;
			trans = (origvalue + pulse);
		end

		if (trans < NUMTRANSMAPS)
			trans = $1 << V_ALPHASHIFT;

			customhud.CustomNum(v,
				startx+40, starty,
				G_TicsToMinutes(displayTime),
				"SANUM", 2,
				flags|trans
			);

			v.draw(
				startx+56, starty,
				colon,
				flags|trans
			);
			customhud.CustomNum(v,
				startx+64, starty,
				G_TicsToSeconds(displayTime),
				"SANUM", 2,
				flags|trans
			);

			v.draw(
				startx+80, starty,
				colon,
				flags|trans
			);
			customhud.CustomNum(v,
				startx+88, starty,
				G_TicsToCentiseconds(displayTime),
				"SANUM", 2,
				flags|trans
			);
		end

		trans = origvalue + (length - pulse);

		if (trans < NUMTRANSMAPS)
			local colonRed = v.cachePatch("SACOLONR");

			trans = $1 << V_ALPHASHIFT;

			customhud.CustomNum(v,
				startx+40, starty,
				G_TicsToMinutes(displayTime),
				"SARNUM", 2,
				flags|trans
			);

			v.draw(
				startx+56, starty,
				colonRed,
				flags|trans
			);
			customhud.CustomNum(v,
				startx+64, starty,
				G_TicsToSeconds(displayTime),
				"SARNUM", 2,
				flags|trans
			);

			v.draw(
				startx+80, starty,
				colonRed,
				flags|trans
			);
			customhud.CustomNum(v,
				startx+88, starty,
				G_TicsToCentiseconds(displayTime),
				"SARNUM", 2,
				flags|trans
			);
		end
	else
		customhud.CustomNum(v,
			startx+40, starty,
			G_TicsToMinutes(displayTime),
			"SANUM", 2,
			flags|V_HUDTRANS
		);

		v.draw(
			startx+56, starty,
			colon,
			flags|V_HUDTRANS
		);
		customhud.CustomNum(v,
			startx+64, starty,
			G_TicsToSeconds(displayTime),
			"SANUM", 2,
			flags|V_HUDTRANS
		);

		v.draw(
			startx+80, starty,
			colon,
			flags|V_HUDTRANS
		);
		customhud.CustomNum(v,
			startx+88, starty,
			G_TicsToCentiseconds(displayTime),
			"SANUM", 2,
			flags|V_HUDTRANS
		);
	end
end

local function sadxRings(v, player)
	if (player.bot)
		return;
	end

	if (maptol & TOL_NIGHTS) or (isFuckingSpecialStage())
		return;
	end

	if not (player and player.valid) return; end

	local startx = 8;
	local starty = 28;

	local ring = v.cachePatch("SARING");

	local flags = V_SNAPTOLEFT|V_SNAPTOTOP|V_PERPLAYER;
	v.draw(startx, starty-1, ring, flags|V_HUDTRANS);

	if (player.rings <= 0)
		local origvalue = v.localTransFlag() >> V_ALPHASHIFT;
		local length = NUMTRANSMAPS - origvalue;
		local pulse = abs(((leveltime/2) % (NUMTRANSMAPS*2)) - NUMTRANSMAPS);
		local trans = 0;

		if (origvalue >= NUMTRANSMAPS)
			return;
		end

		if (origvalue > 0)
			pulse = ($1 * length) / NUMTRANSMAPS;
			trans = (origvalue + pulse);
		end

		if (trans < NUMTRANSMAPS)
			trans = $1 << V_ALPHASHIFT;
			customhud.CustomNum(v,
				startx+24, starty,
				player.rings,
				"SANUM", 3,
				flags|trans
			);
		end

		trans = origvalue + (length - pulse)

		if (trans < NUMTRANSMAPS)
			trans = $1 << V_ALPHASHIFT
			customhud.CustomNum(v,
				startx+24, starty,
				player.rings,
				"SARNUM", 3,
				flags|trans
			);
		end
	else
		customhud.CustomNum(v,
			startx+24, starty,
			player.rings,
			"SANUM", 3,
			flags|V_HUDTRANS
		);
	end
end

local lastValidSkin = {}; -- For intermission

local function sadxLives(v, player)
	if (player.bot)
		return;
	end

	if (maptol & TOL_NIGHTS) or (isFuckingSpecialStage())
		return;
	end

	if not (gametypeFuckingUsesLives()) return; end
	if not (player and player.valid) return; end

	local skin = nil;
	local color = player.skincolor;

	if (player.mo and player.mo.valid)
		skin = player.mo.skin;
		color = player.mo.color;
	elseif (lastValidSkin[#player] != nil)
		skin = lastValidSkin[#player];
	end

	if not (skins[skin] and skins[skin].valid)
		return;
	end

	lastValidSkin[#player] = skin;

	local startx = 8;
	local starty = screenheight-34;

	local bg = v.cachePatch("SAFACEBG");
	local face = v.getSprite2Patch(skin, SPR2_LIFE, (player.powers[pw_super] != 0), 0, 0);

	local flags = V_SNAPTOLEFT|V_SNAPTOBOTTOM|V_PERPLAYER;

	v.draw(
		startx, starty,
		bg,
		flags|V_HUDTRANS
	);
	v.draw(
		(startx+11), (starty+15),
		face,
		flags|V_HUDTRANS,
		v.getColormap(skin, color)
	);
	customhud.CustomNum(v,
		startx+24, starty+5,
		player.lives,
		"SANUM", 2,
		flags|V_HUDTRANS
	);
end

local function sadxBoss(v, player)
	if (player.bot)
		return;
	end

	if (maptol & TOL_NIGHTS) or (isFuckingSpecialStage())
		return;
	end

	if (player.sabossmeter == nil) return; end

	local startx = screenwidth-17;
	local starty = 21;

	local bar = v.cachePatch("SABAR");
	local ends = v.cachePatch("SABAREND");

	local fills = {
		[1] = {
			{22, -5, 31}, -- height, offset, color
			{12, 0, 19},
			{3, -5, 0},
			{3, 13, 0},
			{4, 1, 17},
			{2, 1, 15},
			{2, 9, 21},
			{1, 12, 21}
		},
		[2] = {
			{12, 0, 134},
			{4, 1, 132},
			{2, 1, 130},
			{2, 9, 135},
			{1, 12, 135}
		},
	};

	local leftx = startx - (player.sabossmeter.barlength - 1) - 16;
	local flags = V_SNAPTOTOP|V_SNAPTORIGHT|V_PERPLAYER;

	local masterfill = fills[1];
	for i = 1,#masterfill do
		local curfill = masterfill[i];
		v.drawFill(
			leftx+8, starty+curfill[2],
			player.sabossmeter.barlength, curfill[1],
			curfill[3]|flags
		);
	end

	if (player.sabossmeter.disphp > 0)
		masterfill = fills[2];
		for k = 1,#masterfill do
			local curfill = masterfill[k];
			local aa = player.sabossmeter.disphp+1;

			v.drawFill(
				startx-8-player.sabossmeter.disphp, starty+curfill[2],
				aa, curfill[1],
				curfill[3]|flags
			);
		end
	end

	v.draw(startx, starty, ends, flags|V_FLIP);
	v.draw(leftx, starty, ends, flags);

	local str = player.sabossmeter.name + ":";
	local namex = (startx - v.stringWidth(str, V_ALLOWLOWERCASE));
	if (leftx < namex)
		namex = leftx;
	end

	v.drawString(namex, starty-13, str, flags|V_ALLOWLOWERCASE);
end

local function sadxGammaBonus(v, player)
	if (player.bot)
		return;
	end

	if (maptol & TOL_NIGHTS) or (isFuckingSpecialStage())
		return;
	end

	if (player.mo.gammaVars == nil)
		return;
	end

	for _,bonus in pairs(player.mo.gammaVars.bonusHud) do
		local x = 8 << FRACBITS + (bonus.offset * 6 << FRACBITS);
		local y = 46 << FRACBITS;
		local flags = V_SNAPTOLEFT|V_SNAPTOTOP|V_PERPLAYER;

		local perfectColors = {
			SKINCOLOR_GAMMABONUS1,
			SKINCOLOR_GAMMABONUS3, // these being flipped is the
			SKINCOLOR_GAMMABONUS2, // only reason this is a table
			SKINCOLOR_GAMMABONUS4,
			SKINCOLOR_GAMMABONUS5,
			SKINCOLOR_GAMMABONUS6,
			SKINCOLOR_GAMMABONUS7,
			SKINCOLOR_GAMMABONUS8,
			SKINCOLOR_GAMMABONUS9,
			SKINCOLOR_GAMMABONUS10,
			SKINCOLOR_GAMMABONUS11
		};

		local isPerfect = false;
		local c = SKINCOLOR_NONE;
		if (bonus.i >= 10)
			c = (leveltime % 11);
			isPerfect = true;
		else
			c = SKINCOLOR_GAMMABONUS1 + (bonus.i - 1);
		end

		local str = "";
		if (bonus.time < TICRATE)
			str = tostring(bonus.amt * 100);
		else
			local bonusLabels = {
				"Good!",
				"Nice!",
				"Great!",
				"Jammin'!",
				"Cool!",
				"Radical!",
				"Tight!",
				"Awesome!",
				"Extreme!",
				"Perfect!"
			};
			str = bonusLabels[ bonus.i ];
		end

		local nextx = x;
		for i = 1,str:len() do
			local nextbyte = str:byte(i,i);
			local clr = c;

			if (isPerfect == true)
				clr = perfectColors[ c + 1 ];
				c = ($1 + 1) % 11;
			end

			/*
			local delta = abs( i - bonus.charHop );
			local hop = 0;
			if (delta == 1)
				hop = 2;
			elseif (delta == 0)
				hop = 3;
			end
			*/

			local result = customhud.CustomFontChar(
				v,
				nextx, y,
				nextbyte, "GMSFT",
				flags, FRACUNIT, clr
			);

			nextx = result;
		end
	end
end

local function mapIsHub(map)
	if (rawget(_G, "sugoi") and sugoi.IsHubLevel())
		return true;
	end

	return false;
end

local function sadxTitle(v, player, tick, endtime)
	if (player.bot) return; end
	if not (mapheaderinfo[gamemap].lvlttl) return; end
	if (mapheaderinfo[gamemap].levelflags & LF_NOTITLECARD) return; end
	if (tick < 0 or tick > cardlength) return; end

	local startx = screenwidth/2;
	local starty = 3*screenheight/4;

	local cardnum = 1;

	local color = SKINCOLOR_BLUE;
	if (player and player.valid)
		color = player.skincolor;

		if (player.charflags & SF_MACHINE)
			-- Robots get a grey title (for Gamma)
			color = SKINCOLOR_SILVER;
		end
	end

	if (color == SKINCOLOR_YOGURT
	or color == SKINCOLOR_TAN
	or color == SKINCOLOR_PEACHY
	or color == SKINCOLOR_SUNSET
	or color == SKINCOLOR_APRICOT
	or color == SKINCOLOR_ORANGE
	or color == SKINCOLOR_GOLD
	or color == SKINCOLOR_YELLOW)
		-- skincolors on the light orange/yellow ends of the spectrum get inverted shading!
		-- a tiny detail from sa1 tails that no one except me noticed!
		cardnum = $1+2;
	end

	local trans = 0;
	if (tick < NUMTRANSMAPS)
		trans = (NUMTRANSMAPS-tick) << V_ALPHASHIFT;
	elseif (tick > cardlength-NUMTRANSMAPS)
		trans = (tick - (cardlength-NUMTRANSMAPS)) << V_ALPHASHIFT;
	end

	if not (mapheaderinfo[gamemap].actnum)
		cardnum = $1+1;
	end

	local sub = "ACTION STAGE";
	if (#(mapheaderinfo[gamemap].subttl) > 0)
		sub = string.upper(mapheaderinfo[gamemap].subttl);
	elseif (mapheaderinfo[gamemap].levelflags & LF_WARNINGTITLE)
		sub = "BOSS STAGE";
	elseif (isFuckingSpecialStage())
		sub = "SPECIAL STAGE";
	elseif (maptol & TOL_NIGHTS)
		sub = "NiGHTS STAGE";
	elseif (circuitmap)
		sub = "CIRCUIT STAGE";
	elseif (rawget(_G, "sugoi") and sugoi.IsHubLevel())
		sub = "ADVENTURE FIELD";
	end

	local card = v.cachePatch("SATCARD"..cardnum);

	v.draw(
		startx, starty,
		card,
		V_PERPLAYER|trans,
		v.getColormap(TC_DEFAULT, color)
	);

	if (mapheaderinfo[gamemap].actnum)
		customhud.CustomNum(v,
			startx-88, starty-8,
			mapheaderinfo[gamemap].actnum,
			"SALNUM", nil,
			V_PERPLAYER|trans,
			"center"
		);

		local ttllength = customhud.CustomFontStringWidth(v,
			mapheaderinfo[gamemap].lvlttl,
			"SAFNT"
		);

		if (ttllength > 168)
			if (ttllength > 212)
				local diff = ttllength - (screenwidth-108);
				customhud.CustomFontString(v,
					startx-52-diff, starty-10,
					mapheaderinfo[gamemap].lvlttl,
					"SAFNT",
					V_PERPLAYER|trans,
					"left"
				);
			else
				customhud.CustomFontString(v,
					startx-52, starty-10,
					mapheaderinfo[gamemap].lvlttl,
					"SAFNT",
					V_PERPLAYER|trans,
					"left"
				);
			end
		else
			customhud.CustomFontString(v,
				startx+32, starty-10,
				mapheaderinfo[gamemap].lvlttl,
				"SAFNT",
				V_PERPLAYER|trans,
				"center"
			);
		end

		local sublength = customhud.CustomFontStringWidth(v,
			sub,
			"SASFT"
		);

		if (sublength > 120)
			if (sublength > 214)
				local diff = sublength - (screenwidth-106);
				customhud.CustomFontString(v,
					startx-54-diff, starty+6,
					sub,
					"SASFT",
					V_PERPLAYER|trans,
					"left"
				);
			else
				customhud.CustomFontString(v,
					startx-54, starty+6,
					sub,
					"SASFT",
					V_PERPLAYER|trans,
					"left"
				);
			end
		else
			customhud.CustomFontString(v,
				startx+5, starty+6,
				sub,
				"SASFT",
				V_PERPLAYER|trans,
				"center"
			);
		end
	else
		customhud.CustomFontString(v,
			startx, starty-10,
			mapheaderinfo[gamemap].lvlttl,
			"SAFNT",
			V_PERPLAYER|trans,
			"center"
		);
		customhud.CustomFontString(v,
			startx, starty+6,
			sub,
			"SASFT",
			V_PERPLAYER|trans,
			"center"
		);
	end
end

local function sadxIntermission(v)
	if (gametype != GT_COOP)
		hud.enable("intermissiontally");
		return;
	end

	hud.disable("intermissiontally");
	local player = consoleplayer;

	sadxTime(v, player);
	sadxRings(v, player);
	sadxLives(v, player);

	if not (player and player.valid) return; end
	if (intermissionData == nil) return; end

	local numGroups = 3;
	local groupList = {
		{
			{title = "Time", value = ":"},
			{title = "Rings", value = intermissionData.rings},
		},
		intermissionData.bonusList, -- this is pretty silly but :V
		{
			{title = "Total score", value = intermissionData.total},
		},
	};

	local height = 0;
	for i = 1,numGroups do
		local group = groupList[i];
		local groupLength = #group;

		height = $1 + (24 + ((groupLength-1) * 18));

		if (i < numGroups)
			height = $1 + 6;
		end
	end

	local startY = (screenheight / 2) - (height / 2);
	local x = 32;
	local bgY = startY;

	local bgPatches = {
		v.cachePatch("SABBACK1"),
		v.cachePatch("SABBACK2"),
		v.cachePatch("SABBACK3"),
	};
	local bg = nil;

	-- Draw backers
	for i = 1,numGroups do
		local group = groupList[i];
		local groupLength = #group;

		for j = 1,groupLength do
			if (j == 1)
				bg = bgPatches[1];
			else
				bg = bgPatches[2];
			end

			v.draw(
				x, bgY,
				bg,
				V_TRANSLUCENT
			);

			bgY = $1 + bg.height;
		end

		bg = bgPatches[3];
		v.draw(
			x, bgY,
			bg,
			V_TRANSLUCENT
		);

		bgY = $1 + bg.height + 6;
	end

	-- draw text
	local valuesDrawn = 0;
	bgY = startY;

	for i = 1,numGroups do
		local group = groupList[i];
		local groupLength = #group;

		local textY = bgY + 6;

		for j = 1,groupLength do
			local bonusData = group[j];

			if (j == 1)
				bg = bgPatches[1];
			else
				bg = bgPatches[2];
			end

			customhud.CustomFontString(v,
				x + 16, textY,
				bonusData.title,
				"SABFT",
				0,
				"left"
			);

			if (valuesDrawn < intermissionData.displayValues)
				if (bonusData.value == ":")
					-- Draw the timer instead
					local displayTime = intermissionData.time;
					local colon = v.cachePatch("SACOLON");

					customhud.CustomNum(v,
						x + 192, textY,
						G_TicsToMinutes(displayTime),
						"SANUM", 2,
						0,
						"right"
					);

					v.draw(
						x + 192, textY,
						colon,
						0
					);
					customhud.CustomNum(v,
						x + 216, textY,
						G_TicsToSeconds(displayTime),
						"SANUM", 2,
						0,
						"right"
					);

					v.draw(
						x + 216, textY,
						colon,
						0
					);
					customhud.CustomNum(v,
						x + 240, textY,
						G_TicsToCentiseconds(displayTime),
						"SANUM", 2,
						0,
						"right"
					);
				else
					customhud.CustomNum(v,
						x + 240, textY,
						bonusData.value,
						"SANUM", nil,
						0,
						"right"
					);
				end

				valuesDrawn = $1 + 1;
			end

			textY = $1 + 18;
			bgY = $1 + bg.height;
		end

		bg = bgPatches[3];
		bgY = $1 + bg.height + 6;
	end
end

customhud.SetupItem("score", modname, function(v) end); -- Go away!
customhud.SetupItem("time", modname, sadxTime);
customhud.SetupItem("rings", modname, sadxRings);
customhud.SetupItem("lives", modname, sadxLives);

customhud.SetupItem("stagetitle", modname, sadxTitle);
customhud.SetupItem("intermissiontally", modname, sadxIntermission);

customhud.SetupItem("bossmeter", modname, sadxBoss, "game", 0);
customhud.SetupItem("gammaBonus", modname, sadxGammaBonus, "game", 2);
